/* crypto/pkcs7/sign.c */
/* Copyright (C) 1995-1998 Eric Young (eay@cryptsoft.com)
 * All rights reserved.
 *
 * This package is an SSL implementation written
 * by Eric Young (eay@cryptsoft.com).
 * The implementation was written so as to conform with Netscapes SSL.
 * 
 * This library is free for commercial and non-commercial use as long as
 * the following conditions are aheared to.  The following conditions
 * apply to all code found in this distribution, be it the RC4, RSA,
 * lhash, DES, etc., code; not just the SSL code.  The SSL documentation
 * included with this distribution is covered by the same copyright terms
 * except that the holder is Tim Hudson (tjh@cryptsoft.com).
 * 
 * Copyright remains Eric Young's, and as such any Copyright notices in
 * the code are not to be removed.
 * If this package is used in a product, Eric Young should be given attribution
 * as the author of the parts of the library used.
 * This can be in the form of a textual message at program startup or
 * in documentation (online or textual) provided with the package.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    "This product includes cryptographic software written by
 *     Eric Young (eay@cryptsoft.com)"
 *    The word 'cryptographic' can be left out if the rouines from the library
 *    being used are not cryptographic related :-).
 * 4. If you include any Windows specific code (or a derivative thereof) from 
 *    the apps directory (application code) you must include an acknowledgement:
 *    "This product includes software written by Tim Hudson (tjh@cryptsoft.com)"
 * 
 * THIS SOFTWARE IS PROVIDED BY ERIC YOUNG ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The licence and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution licence
 * [including the GNU Public Licence.]
 */

#include <openca/general.h>

int generatePKCS7 ( int mode, int verbose, char *infile, char *outfile, char *certfile,
                    char *keyfile, char *key, int keyform,
                    char *passargin, int nodetach, char* engine, int is_signed, char *othercertsfile,
                    const EVP_CIPHER *cipher, int flags ) {

	/* BIO *bio_err=NULL; */
	BIO *Sout=NULL;

	/* PKCS7 USED VARIABLES */
	X509 *x509 = NULL;
	STACK_OF(X509) *other = NULL;
	EVP_PKEY *pkey = NULL;
	PKCS7 *p7 = NULL;
	PKCS7_SIGNER_INFO *si = NULL;
	BIO *certf = NULL, *keyf = NULL;
	BIO *data = NULL,*p7bio = NULL;
	BIO *STDout=NULL;
	char buf[1024*4];
	int i;

	/* ADD ENGINE SUPPORT */
	ENGINE *e = NULL;

        if ((bio_err=BIO_new(BIO_s_file())) != NULL)
        	BIO_set_fp(bio_err,stderr,BIO_NOCLOSE|BIO_FP_TEXT);
	data=BIO_new(BIO_s_file());
	STDout=BIO_new_fp(stdout,BIO_NOCLOSE);

	if( infile == NULL ) {
		BIO_set_fp(data,stdin,BIO_NOCLOSE);
	} else {
		if (!BIO_read_filename(data,infile))
			goto err;
		if( verbose )
        		BIO_printf(STDout,"Input file intialized.\n");
	}

	ERR_load_crypto_strings();

	/* ENGINE support added */
#ifdef OPENSSL_ENGINE
	e = (ENGINE *) setup_engine(bio_err, engine, 0);
#endif

	if( OPENCA_PKCS7_SIGN == mode && is_signed ) {
		/* Passin support */
		if(!key && !app_passwd(bio_err, passargin, NULL, &key, NULL)) {
			BIO_printf(bio_err, "Error getting passwords\n");
			goto err;
		}

#ifdef OPENSSL_ENGINE
		pkey = (EVP_PKEY *) load_key(bio_err, keyfile, keyform, 0, key,
			e, "private key");
#else
		pkey = (EVP_PKEY *) load_key(bio_err, keyfile, keyform, 0, key,
			NULL, "private key");
#endif
		if (key) memset(key,0,strlen(key));
		if (pkey == NULL)
			{
			/* load_key() has already printed an appropriate message */
			goto err;
			}
	}

	/* Read Certificate file and certficate data */
	if( verbose )
        	BIO_printf(STDout,"Reading Certificate file.\n");
#ifdef OPENSSL_ENGINE
	if ((x509= (X509 *) load_cert(bio_err, certfile, FORMAT_PEM, NULL, e, "certificate")) == NULL)
#else
	if ((x509= (X509 *) load_cert(bio_err, certfile, FORMAT_PEM, NULL, NULL, "certificate")) == NULL)
#endif
		goto err;
	if (OPENCA_PKCS7_SIGN == mode && !X509_check_private_key(x509,pkey))
		{
		BIO_printf(bio_err,"CA certificate and CA private key do not match\n");
		goto err;
		}

	if (mode == OPENCA_PKCS7_SIGN)
	{
		if( verbose )
        		BIO_printf(STDout,"Start signing.\n");
		p7=PKCS7_new();
	
		if( is_signed == 1 ) {
			PKCS7_set_type(p7,NID_pkcs7_signed);
		 
			si=PKCS7_add_signature(p7,x509,pkey,EVP_sha1());
			if (si == NULL) goto err;
	
			/* Get signing time automatically added */
			PKCS7_add_signed_attribute(si, NID_pkcs9_contentType,
					V_ASN1_OBJECT, OBJ_nid2obj(NID_pkcs7_data));
	
		} else {
			PKCS7_set_type(p7,NID_pkcs7_signed);
			// PKCS7_set_type(p7,NID_pkcs7_enveloped);
		}
	
		/* we may want to add more certificates to the structure, for example
		 * in a scep response */
		if( othercertsfile ) {
			X509 *tmp_cert = NULL;
			char name[1024];
	
			/* don't ask me why we need a cast here */
			other = (STACK_OF(X509) *) load_certs(bio_err, othercertsfile, FORMAT_PEM,
				NULL, e, "other certificates");
	
			if (!other) goto err;
	
			while( sk_X509_num( other ) > 0 ) {
				tmp_cert = sk_X509_pop(other);
	
				PKCS7_add_certificate(p7,tmp_cert);
				if( verbose ) {
					X509_NAME_oneline(X509_get_subject_name(
							tmp_cert ), name, 1024);
					BIO_printf( bio_err, "Addedd certificate %s\n",
							name);
				}
			}
	
		} else {
			/* we may want to add more */
			PKCS7_add_certificate(p7,x509);
		}


		/* Set the content of the signed to 'data' */
		PKCS7_content_new(p7,NID_pkcs7_data);

		if (!nodetach)
			PKCS7_set_detached(p7,1);

		if ((p7bio=PKCS7_dataInit(p7,NULL)) == NULL) goto err;

		if (verbose)
       		 	BIO_printf(STDout,"Reading Data to be signed.\n");
		for (;;)
		{
			i=BIO_read(data,buf,sizeof(buf));
			if (i <= 0) break;
			BIO_write(p7bio,buf,i);
		}

		if (!PKCS7_dataFinal(p7,p7bio)) goto err;
		BIO_free(p7bio);

	} else { /* OPENCA_PKCS7_ENCRYPT */
		if( verbose )
        		BIO_printf(STDout,"Start encrpytion.\n");
		other = sk_X509_new_null ();
		if( verbose )
        		BIO_printf(STDout,"X509 stack initialized.\n");
		sk_X509_push(other, x509);
		if( verbose )
        		BIO_printf(STDout,"X509 stack filled.\n");
		p7 = PKCS7_encrypt(other, data, EVP_aes_256_cbc(), flags | PKCS7_BINARY);
		// p7 = PKCS7_encrypt(other, data, cipher, flags);
		if( verbose )
        		BIO_printf(STDout,"Data encrypted.\n");
        }

	Sout=BIO_new(BIO_s_file());
	if (outfile != NULL)
	{
        	if (BIO_write_filename(Sout,outfile) <= 0)
                {
			BIO_printf(bio_err,"Error writing file %s\n", outfile);
                	perror(outfile);
                        goto err;
                }
        } else {
        	BIO_set_fp(Sout,stdout,BIO_NOCLOSE|BIO_FP_TEXT);
	}

	PEM_write_bio_PKCS7(Sout,p7);
	PKCS7_free(p7);

	return 1;

err:
	ERR_print_errors_fp(stderr);
	exit(1);
	}
